// <copyright file="ArqCs.LoopThread.cs" company="dev.arqendra.net">
// v1.4
// </copyright>
// <author>TrV</author>
// <email>dev(at)arqendra(dot)net</email>
// <date>2010-12</date>
// <summary>Class LoopThread.</summary>

using System;

namespace ArqCs
{
    #region constants
    /// <summary>
    /// Thread state.
    /// </summary>
    public enum ThreadState
    {
        /// <summary>
        /// State 'Created'.
        /// </summary>
        Created,

        /// <summary>
        /// State 'Started'.
        /// </summary>
        Started,

        /// <summary>
        /// State 'Paused'.
        /// </summary>
        Paused,

        /// <summary>
        /// State 'Resumed'.
        /// </summary>
        Resumed,

        /// <summary>
        /// State 'Stopped'.
        /// </summary>
        Stopped,

        /// <summary>
        /// State 'Destroyed'.
        /// </summary>
        Destroyed,

        /// <summary>
        /// State 'Polled'.
        /// </summary>
        ////Polled,

        /// <summary>
        /// State 'Executed'.
        /// </summary>
        ////Executed
    }
    #endregion

    #region ThreadEventArgs
    /// <summary>
    /// Holds information about thread state.
    /// </summary>
    public class ThreadEventArgs : EventArgs
    {
        #region attributes
        /// <summary>
        /// Thread identificator.
        /// </summary>
        private string id;

        /// <summary>
        /// State (Created/Started/Paused/Resumed/Stopped/Destroyed).
        /// </summary>
        private ThreadState state;
        #endregion

        #region properties
        /// <summary>
        /// Gets identificator.
        /// </summary>
        /// <value>Identificator string.</value>
        public string Id
        {
            get { return this.id; }
        }

        /// <summary>
        /// Gets thread state.
        /// </summary>
        /// <value>Thread state.</value>
        public ThreadState State
        {
            get { return this.state; }
        }
        #endregion

        #region constructor
        /// <summary>
        /// Initializes a new instance of the ThreadEventArgs class.
        /// </summary>
        /// <param name="id">Thread identificator.</param>
        /// <param name="state">Thread state.</param>
        public ThreadEventArgs(string id, ThreadState state)
        {
            this.id = id;
            this.state = state;
        }
        #endregion
    }
    #endregion

    #region LoopThread
    /// <summary>
    /// Executes repeated actions in a parallel process.
    /// </summary>
    public abstract class LoopThread
    {
        #region attributes
        /// <summary>
        /// Thread identificator.
        /// </summary>
        private string id;

        /// <summary>
        /// Time between two polls (useful to poll for thread commands).
        /// </summary>
        private int pollTime;

        /// <summary>
        /// Time between two loops (real sleep time).
        /// </summary>
        private int loopTime;

        /// <summary>
        /// Countdown between two loops.
        /// </summary>
        private int countdown;

        /// <summary>
        /// Thread running state.
        /// </summary>
        private volatile bool running;

        /// <summary>
        /// Thread suspended state.
        /// </summary>
        private volatile bool suspended;
        #endregion

        #region properties
        /// <summary>
        /// Gets or sets poll time.
        /// </summary>
        /// <value>Poll time [milliseconds].</value>
        /// <exception cref="InvalidOperationException">Poll time has been modified while thread is running.</exception>
        protected int PollTime
        {
            get {
                return this.pollTime;
            }

            set {
                if ( !this.running ) { // if thread is not running
                    this.pollTime = value;
                }
                else {
                    throw new System.InvalidOperationException("Poll time cannot be modified while thread is running");
                }
            }
        }

        /// <summary>
        /// Gets or sets loop time.
        /// </summary>
        /// <value>Loop time [milliseconds].</value>
        /// <exception cref="InvalidOperationException">Loop time has been modified while thread is running.</exception>
        protected int LoopTime
        {
            get {
                return this.loopTime;
            }

            set {
                if ( !this.running ) { // if thread is not running
                    this.loopTime = value;
                }
                else {
                    throw new System.InvalidOperationException("Loop time cannot be modified while thread is running");
                }
            }
        }

        /// <summary>
        /// Gets or sets identificator.
        /// </summary>
        /// <value>Identificator string.</value>
        /// <exception cref="InvalidOperationException">Identificator has been modified while thread is running.</exception>
        public string Id
        {
            get {
                return this.id;
            }

            set {
                if ( !this.running ) { // if thread is not running
                    this.id = value;
                }
                else {
                    throw new System.InvalidOperationException("Identificator cannot be modified while thread is running");
                }
            }
        }

        /// <summary>
        /// Gets a value indicating whether thread is running or not.
        /// </summary>
        /// <value>True/False if thread is/is not running.</value>
        public bool Running
        {
            get { return this.running; }
        }

        /// <summary>
        /// Gets a value indicating whether thread is suspended or not.
        /// </summary>
        /// <value>True/False if thread is/is not suspended.</value>
        public bool Suspended
        {
            get { return this.suspended; }
        }
        #endregion

        #region constructors
        /// <summary>
        /// Initializes a new instance of the LoopThread class.
        /// </summary>
        /// <param name="pollTime">Time between two polls (useful to poll for thread commands).</param>
        /// <param name="loopTime">Time between two loops (real sleep time).</param>
        /// <exception cref="ArgumentOutOfRangeException">Poll time or loop time is lower or equal to 0.</exception>
        /// <exception cref="ArgumentException">Poll time is greater than loop time.</exception>
        protected LoopThread(int pollTime, int loopTime)
            : this(String.Empty, pollTime, loopTime)
        {
        }

        /// <summary>
        /// Initializes a new instance of the LoopThread class.
        /// </summary>
        /// <param name="id">Thread identificator.</param>
        /// <param name="pollTime">Time between two polls (useful to poll for thread commands).</param>
        /// <param name="loopTime">Time between two loops (real sleep time).</param>
        /// <exception cref="ArgumentOutOfRangeException">Poll time or loop time is lower or equal to 0.</exception>
        /// <exception cref="ArgumentException">Poll time is greater than loop time.</exception>
        protected LoopThread(string id, int pollTime, int loopTime)
        {
            if ( pollTime <= 0 ) {
                throw new System.ArgumentOutOfRangeException("pollTime", "Poll time cannot be lower or equal to 0");
            }

            if ( loopTime <= 0 ) {
                throw new System.ArgumentOutOfRangeException("loopTime", "Loop time cannot be lower or equal to 0");
            }

            if ( pollTime > loopTime ) {
                throw new System.ArgumentException("Poll time cannot be greater than loop time");
            }

            this.id = id;
            this.pollTime = pollTime;
            this.loopTime = loopTime;

            this.running = false;
            ////this.change = false;
        }
        #endregion

        #region events
        /// <summary>
        /// Occurs when thread state has changed.
        /// </summary>
        public event EventHandler<ThreadEventArgs> ThreadStateChanged;
        #endregion

        #region event methods
        /// <summary>
        /// Sends event about thread state change.
        /// </summary>
        /// <param name="state">Thread state.</param>
        private void ChangeState(ThreadState state)
        {
            if ( this.ThreadStateChanged != null ) { // if any event handler is attached to the event
                this.ThreadStateChanged(this, new ThreadEventArgs(this.id, state)); // sends event
            }
        }
        #endregion

        #region management methods
        /// <summary>
        /// Forces loop to execute as soon as possible (on next thread poll).
        /// </summary>
        public void ForceLoopNow()
        {
            this.countdown = 0; // re-initializes countdown
        }

        /// <summary>
        /// Asks the thread to terminate gracefully as soon as possible (on next thread poll).
        /// </summary>
        public void Stop()
        {
            this.running = false;
        }

        /// <summary>
        /// Asks the thread to suspend as soon as possible (on next thread poll).
        /// </summary>
        public void Suspend()
        {
            if ( !this.suspended ) { // if thread is not suspended
                this.suspended = true;
                this.ChangeState(ThreadState.Paused);
            }
        }

        /// <summary>
        /// Asks the thread to resume as soon as possible (on next thread poll).
        /// </summary>
        public void Resume()
        {
            if ( this.suspended ) { // if thread is suspended
                this.suspended = false;
                this.ChangeState(ThreadState.Resumed);
            }
        }
        #endregion

        #region abstract methods
        /// <summary>
        /// What the thread should do once before loops begin.
        /// </summary>
        protected abstract void DoFirst();

        /// <summary>
        /// What the thread should do repeatedly (= in loop).
        /// </summary>
        protected abstract void DoLoop();

        /// <summary>
        /// What the thread should do once after loops end.
        /// </summary>
        protected abstract void DoLast();
        #endregion

        #region overriden methods
        /// <summary>
        /// Stands for current object.
        /// </summary>
        /// <returns>Information about current object.</returns>
        public override string ToString()
        {
            string assemblyName = this.GetType().Namespace + "." + this.GetType().Name;
            if ( !String.IsNullOrEmpty(this.id) ) {
                assemblyName += " - " + this.id;
            }

            return assemblyName;
        }
        #endregion

        #region main method
        /// <summary>
        /// Executes thread work.
        /// </summary>
        public void Run()
        {
            this.ChangeState(ThreadState.Created);

            this.DoFirst(); // executes first actions

            this.countdown = 0; // initializes countdown (ensures actions will be executed the first time at thread start)

            this.ChangeState(ThreadState.Started);

            this.running = true;

            while ( this.running ) { // while thread is active
                if ( !this.suspended ) { // if thread is not suspended
                    if ( this.countdown <= 0 ) { // if countdown has expired
                        this.DoLoop(); // runs thread
                        this.countdown = this.loopTime - this.pollTime; // resets countdown
                    }
                    else { // if countdown has not expired
                        this.countdown -= this.pollTime; // decrements countdown (keeps sleeping)
                    }
                }

                System.Threading.Thread.Sleep(this.pollTime); // sleeps until next poll
            }

            this.ChangeState(ThreadState.Stopped);

            this.DoLast(); // executes last actions

            this.ChangeState(ThreadState.Destroyed);
        }
        #endregion
    }
    #endregion
}
